import threading

import torch
import clip
import requests
from PIL import Image
from io import BytesIO
from typing import Optional, Tuple
import numpy as np
import logging


class FeatureExtractor:
    # __model = None
    # __preprocess = None
    # __device = "ViT-B/32"
    __instance = None
    __lock = threading.Lock()
    __logger = logging.getLogger(__name__)

    def __new__(cls, device = "cuda" if torch.cuda.is_available() else "cpu", model_name = "ViT-B/32"):
        # 第一次检查 - 不带锁
        if cls.__instance is None:
            # 只有在可能需要创建实例时才获取锁
            with cls.__lock:
                # 第二次检查 - 带锁
                if cls.__instance is None:
                    print(f"创建并初始化 CLIP 模型: {model_name} 在设备: {device}")
                    # 创建实例
                    cls.__instance = super().__new__(cls)
                    # 在这里直接完成初始化
                    cls.__instance.__model, cls.__instance.__preprocess = clip.load(model_name, device=device)
                    cls.__instance.__device = device
        return cls.__instance

    @staticmethod
    def resize_with_padding(img, target_size = (224, 224)):
        """
        调整图像大小，保持纵横比并添加填充

        Args:
            img: 输入图像
            target_size: 目标尺寸，默认为(224, 224)

        Returns:
            调整大小后的图像
        """
        # 计算调整大小的比例
        ratio = min(target_size[0] / img.width, target_size[1] / img.height)
        new_size = (int(img.width * ratio), int(img.height * ratio))

        # 调整图像大小，保持长宽比
        resized_img = img.resize(new_size, Image.LANCZOS)

        # 创建新的填充图像
        new_img = Image.new("RGB", target_size, (255, 255, 255))

        # 计算粘贴位置（居中）
        paste_position = ((target_size[0] - new_size[0]) // 2,
                          (target_size[1] - new_size[1]) // 2)

        # 粘贴调整后的图像
        new_img.paste(resized_img, paste_position)

        return new_img

    def extract_from_url(self, image_url):
        """
        从URL加载图像并提取特征向量

        Args:
            image_url: 图像URL

        Returns:
            特征向量，如果提取失败则返回None
        """

        try:
            # 下载图片
            response = requests.get(image_url, stream=True, timeout=10)
            response.raise_for_status()  # 确保请求成功

            # 将图片数据转换为 PIL Image 对象
            image = Image.open(BytesIO(response.content)).convert("RGB")

            return self.extract_from_image(image)

        except requests.RequestException as e:
            self.__logger.error(f"Network error when downloading image from {image_url}: {e}")
            return None
        except Exception as e:
            self.__logger.error(f"Error extracting features from URL {image_url}: {e}")
            return None

    def extract_from_image(self, img):
        """
        从PIL图像对象提取特征向量

        Args:
            img: 输入图像

        Returns:
            特征向量，如果提取失败则返回None
        """

        try:
            # 调整图像大小并添加填充
            image = self.resize_with_padding(img)

            # 预处理并提取特征
            image_tensor = self.__preprocess(image).unsqueeze(0).to(self.__device)

            with torch.no_grad():
                image_features = self.__model.encode_image(image_tensor)
                # 归一化特征向量
                image_features /= image_features.norm(dim=-1, keepdim=True)

            return image_features.cpu().numpy().flatten()

        except Exception as e:
            self.__logger.error(f"Error extracting features from image: {e}")
            return None